package ch.sahits.datastructure;

import java.util.Comparator;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
/**
 * This class implements a sorted map with limited entries.
 * The map is sorted by its keys. When elements are added
 * to the map it is ensured that
 * <pre>this.size()<=this.listLimit()</pre>
 * is always true.<br/>
 * When an element is added and this results in a violation
 * of the above invariant, the last element of the map is removed
 * to restore the invariant.
 * @author Andi Hotz, Sahits GmbH
 * @since 1.2.4
 *
 * @param <K> Key element of the map. Any key element must be mutably comparable
 * @param <V> Value  type stored in the map
 */
public class LimitedSortedMap<K, V> extends TreeMap<K, V>{

	/**
	 * 
	 */
	private static final long serialVersionUID = -3291012713401815134L;
	private final int listLimit;
	/**
	 * Constructor initializing the limit of the map
	 * @param limit of the number of elements in the map.
	 */
	public LimitedSortedMap(int limit) {
		super();
		listLimit=limit;
	}
	/**
	 * Constructs a new, empty map, ordered according to the given comparator. 
	 * All keys inserted into the map must be mutually comparable by the given 
	 * comparator: comparator.compare(k1, k2) must not throw a ClassCastException for any 
	 * keys k1 and k2 in the map. If the user attempts to put a key into the map that 
	 * violates this constraint, the put(Object key, Object value) call will throw a 
	 * ClassCastException.
	 * @param comparator {@inheritDoc}
	 * @param limit of the number of elements in the map.
	 */
	public LimitedSortedMap(Comparator<? super K> comparator,int limit) {
		super(comparator);
		listLimit=limit;
	}
	/**
	 * {@inheritDoc}
	 * @param m {@inheritDoc}
	 * @param limit of the number of elements in the map.
	 */
	public LimitedSortedMap(Map<? extends K, ? extends V> m,int limit) {
		super(m);
		listLimit=limit;
		reduceToListLimit();
	}

	/**
	 * {@inheritDoc}
	 * @param m {@inheritDoc}
	 * @param limit of the number of elements in the map.
	 */
	public LimitedSortedMap(SortedMap<K, ? extends V> m,int limit) {
		super(m);
		listLimit=limit;
		reduceToListLimit();
	}
	/**
	 * Reduce the size of the map to listLimit
	 */
	private void reduceToListLimit() {
		while (size()>listLimit){
			remove(this.lastKey());
		}
	}
	/**
	 * Put a new key value pair into the map and
	 * return the value.
	 * If the key would be inserted after the last element
	 * and the limit of the map size is reached
	 * the key value pair will not be inserted and
	 * the returned value is null
	 */
	@Override
	public V put(K key, V value) {
		super.put(key, value);
		reduceToListLimit();
		return get(key);
	}
	/**
	 * Add a map to this map. The elements
	 * are added in order. 
	 * If the key of an element would be inserted after the last element
	 * and the limit of the map size is reached
	 * that key value pair will not be inserted
	 */
	@Override
	public void putAll(Map<? extends K, ? extends V> map) {
		super.putAll(map);
		reduceToListLimit();
	}
	/**
	 * Retrieve the limit of the map size
	 * @return map limit
	 */
	public final int listLimit(){
		return listLimit;
	}
	
	
	
	
}
