package ch.sahits.test;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.HashSet;
import java.util.Set;

/**
 * This file comparator can compare to text based
 * files.
 * @author Andi Hotz, Sahits GmbH
 * @since 2.1.0
 */
public final class FileComparator {
	/** File accessor for the file to be compared to */
	private final RandomAccessFile raf;
	private final boolean javafile;
	/**
	 * Initialize this instance with the instance with the file to be compared to
	 * @param fileName absolute path to the file
	 * @throws FileNotFoundException fileName was not a valid path 
	 */
	public FileComparator(String fileName) throws FileNotFoundException{
		raf = new RandomAccessFile(new File(fileName), "r");
		javafile=fileName.endsWith(".java");
	}
	/**
	 * Compare the file denoted by <code>fileName</code> to the file initialized in
	 * the constructor.
	 * The comparison ignores empty lines. Multiple spaces and tabs are handled as a
	 * single space.
	 * Tow textual files are equal under the above assumptions when every line of the
	 * file is equal.
	 * @param fileName absolute path of the file to compare this to
	 * @return Instance that holds the result of the comparison
	 * @throws FileNotFoundException fileName was not a valid path 
	 * @throws IOException Error during the reading of the file
	 */
	public ComparetorResult equals(String fileName) throws IOException{
		RandomAccessFile innerRaf=new RandomAccessFile(new File(fileName), "r");
		String l1 = getNextLine(raf);
		String l2 = getNextLine(innerRaf);
		while (l1!=null && l2!=null){
			if (l1==null && l2!=null){
				return new ComparetorResult(false, fileName+" is longer than this file");
			}
			if (l1!=null && l2==null){
				return new ComparetorResult(false, fileName+" is shorter than this file");
			}
			// skip newlines
			while (l1.equals("") && !l2.equals("")){
				l1=getNextLine(raf);
				if (l1==null){
					return new ComparetorResult(false, fileName+" is longer than this file");
				}
			}
			while (!l1.equals("") && l2.equals("")){
				l2=getNextLine(innerRaf);
				if (l2==null){
					return new ComparetorResult(false, fileName+" is shorter than this file");
				}
			}
			if (areImports(l1, l2)){
				Set<String> imports1 = new HashSet<String>();
				Set<String> imports2 = new HashSet<String>();
				while (areImports(l1,l2)){
					imports1.add(l1);
					imports2.add(l2);
					l1 = getNextLine(raf);
					l2 = getNextLine(innerRaf);
				}
				if (!imports1.equals(imports1)){
					return new ComparetorResult(false, "The import sets of the two files differ.");
				}
			}
			if (!l1.equals(l2)){
				// the two lines are not equal
				return new ComparetorResult(false, "The line '"+l1+"' differs from '"+l2+"'");
			}
			l1 = getNextLine(raf);
			l2 = getNextLine(innerRaf);
		}
		// l1 and l2 are both null => EOF reached
		if (l1==null && l2!=null){
			return new ComparetorResult(false, fileName+" is longer than this file");
		}
		if (l1!=null && l2==null){
			return new ComparetorResult(false, fileName+" is shorter than this file");
		}
		return ComparetorResult.TRUE;
	}
	/**
	 * Check if the two lines are import statements
	 * @param l1 line from file 1
	 * @param l2 line from file 2
	 * @return true if the file is a java file and both lines are import statements
	 */
	private boolean areImports(String l1, String l2) {
		if (javafile){
			if (l1.startsWith("import") && l2.startsWith("import")){
				return true;
			}
		}
		return false;
	}
	/**
	 * Retrieve the next unread line of the passed file accessor
	 * @param _raf File accessor to be used
	 * @return next unread line or null if the end of file is reached
	 * @throws IOException Error during the reading of the file
	 */
	private String getNextLine(RandomAccessFile _raf) throws IOException {
		String s = _raf.readLine();
		if (s==null) return null;
		s = s.replace('\t', ' ');
		s=s.trim();
		s=s.replaceAll("\n\r", "\n");
		while (s.indexOf("  ")>=0){
			s = s.replaceAll("  ", " ");
		}
		return s;
	}
}
