package ch.sahits.test;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.util.StringTokenizer;
/**
 * This class is a tokenizer for a text file. Internaly the tokenizer uses a StringTokenizer.
 * Line endings '\n' and '\r' are converted into spaces.
 * @author Andi Hotz, Sahits GmbH
 */
public class TextFileTokenizer extends StringTokenizer {
	/**
	 * Initialize the tokenizer with delimiters and delimiter flag
	 * @param fileName fully qualified path to the file to be read
	 * @param delim {@inheritDoc}
	 * @param returnDelims {@inheritDoc}
	 * @throws FileNotFoundException file path is invalid
	 * @throws IOException Error while reading from the file
	 */
	public TextFileTokenizer(String fileName, String delim, boolean returnDelims) throws IOException {
		super(file2String(fileName), delim, returnDelims);
	}
	/**
	 * Initialize the tokenizer with delimiters
	 * @param fileName fully qualified path to the file to be read
	 * @param delim {@inheritDoc}
	 * @throws FileNotFoundException file path is invalid
	 * @throws IOException Error while reading from the file
	 */
	public TextFileTokenizer(String fileName, String delim) throws IOException {
		super(file2String(fileName), delim);
	}
	/**
	 * Initialize the tokenizer with delimiters
	 * @param fileName fully qualified path to the file to be read
	 * @throws FileNotFoundException file path is invalid
	 * @throws IOException Error while reading from the file
	 */
	public TextFileTokenizer(String fileName) throws IOException {
		super(file2String(fileName));
	}
	/**
	 * Create a tokenizer based on an input stream.
	 * The stream is not closed after it is passed on
	 * @param in Input Stream
	 * @param delim {@inheritDoc}
	 * @throws IOException Error while reading from the stream
	 */
	public TextFileTokenizer(InputStream in, String delim) throws IOException{
		super(stream2String(in),delim);
	}
	/**
	 * Convert the text file into a string. The line terminations are replaced by spaces
	 * @param fileName fully qualified path to the file to be read
	 * @return String of the file contents
	 * @throws FileNotFoundException file path is invalid
	 * @throws IOException Error while reading from the file
	 */
	private static String file2String(String fileName) throws IOException{
		RandomAccessFile raf = new RandomAccessFile(fileName, "r");
		String s =raf.readLine();
		StringBuilder builder = new StringBuilder();
		while (s!=null){
			if (!s.endsWith(" ")){
				// only append a space to replace Line delimiters if no space was already present
				builder.append(s+" ");
			} else {
				builder.append(s);
			}
			s = raf.readLine();
		}
		return builder.toString();
	}
	/**
	 * Read the file from a stream
	 * @param in InputStream
	 * @return Contents of the file as a string
	 * @throws IOException Error while reading
	 */
	private static String stream2String(InputStream in) throws IOException{
		StringBuilder builder = new StringBuilder();
		int ch;
		int pos=0; // current position in th builder
		while ((ch=in.read())!=-1){
			char c = (char)ch;
			if (c=='\n'){
				c=' ';
			}
			if (c=='\r'){
				c=' ';
			}
			if (pos>0 && builder.charAt(pos-1)==' ' && c==' '){
				// do not append multiple spaces
			} else {
				pos++;
				builder.append(c);
			}
		}
		return builder.toString();
	}


}
