package ch.sahits.util;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.List;
import java.util.Vector;

/**
 * This class provides method to check whether
 * some interface is implemented by the class
 * or the class is the subclass of any class.
 * @author Andi Hotz
 *
 */
public class ClassChecker {
	/** Class to be checked */
	private Class<? extends Object> checkClass=null;

	/**
	 * Initialize the class to be checked
	 * @param checkClass
	 */
	@SuppressWarnings("unchecked")
	public ClassChecker(Class checkClass) {
		super();
		this.checkClass = checkClass;
	}
	/**
	 * Check whether the class implements a interface.
	 * The interface must be declared as implemented in
	 * the class or the super class or the super class of
	 * the super class ...
	 * @param interfaceClass that should be implemented
	 * @return true if the interface is anywhere in the class hierarchy implemented.
	 */
	public boolean implementsInterface(Class<? extends Object> interfaceClass){
		Class<? extends Object>[] interfaces = checkClass.getInterfaces();
		for (int i = 0; i < interfaces.length; i++) {
			//if (interfaces[i].equals(interfaceClass)){
			if (checkInterface(interfaceClass,interfaces[i])){
				return true;
			}
		}
		Class<? extends Object> superClass = checkClass.getSuperclass();
		if (superClass==null || superClass.equals(Object.class)){
			// Recursion stops here
			return false;
		} else {
			return new ClassChecker(superClass).implementsInterface(interfaceClass);
		}
	}
	/**
	 * check if the interface is implemented through the class itself, its superclass or any of its
	 * super interfaces (hierarchically)
	 * @param interfaceClass
	 * @return
	 */
	public boolean implementsInterfaceIndirectly(Class<? extends Object> interfaceClass){
		if (implementsInterface(interfaceClass)){
			return true;
		}
		Class[] superInterfaces = checkClass.getInterfaces();
		if (superInterfaces.length>0){
			for (Class clazz : superInterfaces) {
				ClassChecker cc = new ClassChecker(clazz);
				if (cc.implementsInterfaceIndirectly(interfaceClass)){
					return true;
				}
			}
		}
		return false;
	}
	/**
	 * Check whether the <code>test</code> interface is or extends the
	 * <code>reference</code> interface
	 * @param reference interface to compare to
	 * @param test interface to be checked
	 * @return true if test equals reference or the same is true for the super interface
	 */
	private boolean checkInterface(Class<? extends Object> reference, Class<? extends Object> test){
		if (!reference.isInterface() || !test.isInterface()){
			return false;
		}
		if (reference.equals(test)){
			return true;
		}
		// It is possible that the classes where loaded by different class loaders and are therefore unequal
		if (reference.getName().equals(test.getName())){
			// If the interfaces have the same name but are not the same there is something fundamently wrong
			return true;
		}
		Class<? extends Object>[] superInterfaces = test.getInterfaces();
		boolean found = false;
		for (int i = 0; i < superInterfaces.length; i++) {
			found= checkInterface(reference, superInterfaces[i]);
			if (found){
				return true;
			}
		}
		return false;
	}
	/**
	 * Check whether the class extends a class.
	 * The super class must be declared as extending the
	 * class or the super class or the super class ...
	 * @param superClass that should be extended
	 * @return true if the class is extending the super class
	 */
	public boolean extendsClass(Class<? extends Object> superClass){
		Class<? extends Object> c =checkClass.getSuperclass();
		if (c.equals(Object.class)){
			// Recursion stops here
			return false;
		}
		if (c.equals(superClass)){
			return true;
		} else {
			return new ClassChecker(c).extendsClass(superClass);
		}
	}
	/**
	 * Load the class by its name and call {@link #extendsClass(Class)}
	 * @param superClassName name of the super class
	 * @return false if the superclass could not be loaded or the result from {@link #extendsClass(Class)}
	 */
	public boolean extendsClass(String superClassName){
		try {
			return extendsClass(getClass().getClassLoader().loadClass(superClassName));
		} catch (ClassNotFoundException e) {
			return false;
		}
	}
	/**
	 * Load the class by its name and call {@link #implementsInterface(Class)}
	 * @param interfaceName name of the interface
	 * @return false if the interface could not be loaded or the result from {@link #implementsInterface(Class)}
	 */
	public boolean implementsInterface(String interfaceName){
		try {
			return implementsInterface(getClass().getClassLoader().loadClass(interfaceName));
		} catch (ClassNotFoundException e) {
			return false;
		}
	}
	/**
	 * Check if the class implements the specified interface directly and
	 * not through another interface or a super class.
	 * @param intervace interface to be checked
	 * @return true if the class implements the interface directly
	 * @throws IllegalArgumentException if the passed interface is not an interface at all.
	 */
	public boolean implementsDirectly(Class intervace){
		if (!intervace.isInterface()){
			throw new IllegalArgumentException("The passed class "+intervace.getName()+" is not an interface");
		}
		Class<?>[] interfaces = checkClass.getInterfaces();
		for (Class<?> clazz : interfaces) {
			if (clazz==intervace){
				return true;
			}
		}
		return false;
	}
	/**
	 * Check whether the class implements a constructor with
	 * the given types. Constructors that take a more general
	 * type as arguments are also accepted as long they extend
	 * the type explicitly. This excludes the type Object.
	 * @param types Array of argument types
	 * @return true if such a constructor exists
	 */
	public boolean hasConstructor(Class<? extends Object>[] types){
		Constructor<? extends Object>[] conts = checkClass.getConstructors();
		boolean notOk = true;
		for (int i = 0; i < conts.length; i++) {
			Class<? extends Object>[] args = conts[i].getParameterTypes();
			if (args.length!=types.length){
				continue;	// not the same number of arguments
			}
			boolean typesOk = true;
			// check for identity
			for (int j = 0; j < args.length; j++) {
				if (!args[j].equals(types[j])){
					typesOk=false; // one type is not identical
					break;
				}
			}
			if (typesOk){
				notOk=false;
			}
			// check if the types in args are more general than the one in types
			for (int j = 0; j < args.length; j++) {
				ClassChecker cc = new ClassChecker(args[j]);
				if (types[j].isInterface()){
					// args[j] may implement the type
					if (!cc.implementsInterface(types[j])){
						typesOk=false;
						break;
					}
				} else {
					if (cc.extendsClass(types[j])){
						typesOk=false;
						break;
					}
				}
			}
			if (typesOk){
				notOk=false;
			}
		}
		if (notOk) return false;
		else return true;
	}
	/**
	 * Check whether the class implements a method with the given name and
	 * types. Constructors that take a more general
	 * type as arguments are also accepted as long they extend
	 * the type explicitly. This excludes the type Object.
	 * @param name of the method
	 * @param types Array of argument types
	 * @return true if such a method exists
	 */
	public boolean hasMethod(String name, Class<? extends Object>[] types){
		Method[] temp = checkClass.getMethods();
		Vector<Method> v = new Vector<Method>();
		for (int i = 0; i < temp.length; i++) {
			if (temp[i].getName().equals(name)){
				v.add(temp[i]);
			}
		}
		Method[] ms = new Method[v.size()];
		v.copyInto(ms);
		boolean notOk = true;
		for (int i = 0; i < ms.length; i++) {
			Class<? extends Object>[] args = ms[i].getParameterTypes();
			if (args.length!=types.length){
				continue;	// not the same number of arguments
			}
			boolean typesOk = true;
			// check for identity
			for (int j = 0; j < args.length; j++) {
				if (!args[j].equals(types[j])){
					typesOk=false; // one type is not identical
					break;
				}
			}
			if (typesOk){
				notOk=false;
			}
			// check if the types in args are more general than the one in types
			for (int j = 0; j < args.length; j++) {
				ClassChecker cc = new ClassChecker(args[j]);
				if (types[j].isInterface()){
					// args[j] may implement the type
					if (!cc.implementsInterface(types[j])){
						typesOk=false;
						break;
					}
				} else {
					if (cc.extendsClass(types[j])){
						typesOk=false;
						break;
					}
				}
			}
			if (typesOk){
				notOk=false;
			}
		}
		if (notOk) return false;
		else return true;		
	}
	/**
	 * Retrieve a list of all super classes of the class to be checked.
	 * The Object class is excluded.
	 * @return List of super classes
	 */
	public List<Class<? extends Object>> superClasses(){
		Vector<Class<? extends Object>> v = new Vector<Class<? extends Object>>();
		Class<? extends Object> c =checkClass.getSuperclass();
		while (!c.equals(Object.class)){
			v.add(c);
			c=c.getSuperclass();
		}
		return v;
	}
	


}
