package ch.sahits.util;

import java.util.Iterator;
import java.util.List;


/**
 * This class helps to ensure, that a call of a
 * method only stems from a specific class
 * or even a specific method within that class.<br>
 * This is a realization of the friend pattern
 * @author Andi Hotz (c) Sahits.ch 2007
 * @version 1.0
 */
public class FriendPattern {
	/**
	 * Flag indicates  if the call may origin
	 * from any class in the class hierarchy.
	 * This means that if B extends A and method foo
	 * is defined on both the call of {@link #checkAccess()}
	 * for foo on class B yields no exception even when the
	 * {@link #originClass} is A.
	 */
	private boolean includeClassHirachy;
	/**
	 * Class from which access is allowed
	 */
	private Class<? extends Object> originClass;
	/** The class to be checked against, this is needed for hierarchical checks */
	private Class<? extends Object> checkClass=null;
	/**
	 * Method from which access is allowed
	 */
	private String methodName=null;
	/**
	 * Constructor initializes the allowed class and method.
	 * The method name may be null if access from all methods
	 * of the indicated class is granted.
	 * @param c Class from whence access is allowed
	 * @param mName method name from which access is allowed.
	 */
	public FriendPattern(Class<? extends Object> c, String mName){
		init(c,mName,false);
	}
	/**
	 * Constructor initializes the allowed class and method.
	 * The method name may be null if access from all methods
	 * of the indicated class is granted.
	 * @param c Class from whence access is allowed
	 * @param mName method name from which access is allowed.
	 * @param includClassHirarchy flag indicating if the class hierarchy should be taken into account
	 * @param class2BChecked Class to be checkted may be null if <code>includClassHirarchy</code> is false
	 */
	public FriendPattern(Class<? extends Object> c, String mName,boolean includClassHirarchy,Class<? extends Object> class2BChecked){
		init(c, mName, includClassHirarchy);
		if (includClassHirarchy){
			checkClass=class2BChecked;
		} else {
			throw new RuntimeException("The class to be checked must be supplied when definig hirarchical checks");
		}
	}
	/**
	 * Initialize the members
	 * @param c Class from whence access is allowed
	 * @param mName method name from which access is allowed.
	 * @param includClassHirarchy flag indicating if the class hierarchy should be taken into account
	 */
	private void init(Class<? extends Object> c, String mName, boolean includClassHirarchy) {
		originClass = c;
		methodName= mName;
		this.includeClassHirachy=includClassHirarchy;
	}
	/**
	 * Check if the method call stems from an allowed Class/Method.
	 * If the call comes from somewhere else an IllegalAccessException
	 * is thrown.<br>
	 * ATTENTION: When in the calling class  more than one method with the allowed
	 * name exist but only form one of them the call is allowed, this case
	 * is not handled here.
	 * @throws RuntimeException Method is not allowed to be accessed or the method was called from somewhere not allowed
	 */
	public void checkAccess(){
		try{
			throw new Exception();
		} catch (Exception ex){
			StackTraceElement ste = ex.getStackTrace()[2];
			// Element at index 0: this method
			// Element at index 1: method from whence the the access is checked
			// Element at index 2: method that's access should be checked
			String className = ste.getClassName();
			if (!className.equals(originClass.getName()) && !includeClassHirachy){
				throw new RuntimeException("This method may only be called from a "+originClass.getName()+" object but not from "+ste.getClassName());
			} else if (includeClassHirachy){
				if (!className.equals(originClass.getName())){
					// Check the hierarchy
					List<Class<? extends Object>> superClasses = new ClassChecker(checkClass).superClasses();
					for (Iterator<Class<? extends Object>> iterator = superClasses.iterator(); iterator.hasNext();) {
						Class<? extends Object> c =  iterator.next();
						FriendPattern fp = new FriendPattern(c,methodName); // don't check in the superclasses that would result in a stack overflow
						try {
							fp.checkAccess();
							// The superclass defines the method
							break;
						} catch (RuntimeException e) {
							// The tested super class does not fullfill the requirement test the next super class
						}
						
					}
				}
			}
			if (methodName!=null){
				if (!ste.getMethodName().equals(methodName)){
					throw new RuntimeException("This method may only be called from "+methodName+" but not from "+ste.getMethodName());
				}
			}
		}
	}
	/**
	 * Check if <code>c</code> is an extension of the originClass
	 * @param c Class to be checked
	 * @return true if <code>originClass</code> is super class of <code>c</code>
	 */
	private boolean isSubClass(Class<? extends Object> c){
		ClassChecker cc = new ClassChecker(c);
		return cc.extendsClass(originClass);
	}
//	/**
//	 * Check if <code>className</code> is an extension of the originClass. If the class for
//	 * className cannot be loaded false is returned
//	 * @param className Class to be checked
//	 * @return true if <code>originClass</code> is super class of <code>className</code>
//	 */
//	private boolean isSubClass(String className){
//		try {
//			return isSubClass(getClass().getClassLoader().loadClass(className));
//		} catch (ClassNotFoundException e) {
//			return false;
//		}
//	}
	/**
	 * Set the ClassLoader that is needed in {@link #checkAccess()}.
	 * This implementation does nothing since the default class loader is
	 * always used.
	 * @param cl Instance of a class loader or null
	 */
	protected void setClassLoader(ClassLoader cl){
		// do nothing - always use the default class loader
	}
	/**
	 * Retrieve the default class loader for this class
	 * @return class loader for this class
	 * @see FriendPattern#getClassLoader()
	 */
	protected ClassLoader getClassLoader(){
		return getClass().getClassLoader();
	}
}
